@@ -1,21 +1,24 @@
-"""Implements the Context and Command classes.
+"""Implements the CmdContext and Command classes.
+
+- A CmdContext creates new command objects. You should use the `context`
+function to create these, rather than creating them directly.
 
-- A Context creates new command objects.
 - A Command specifies the arguments and options used to run a program.
 """
 
-import asyncio
 import dataclasses
 import enum
 import os
 import signal
+from collections.abc import Iterable
 from dataclasses import dataclass, field
 from typing import Any, Optional, TypeVar, Union
 
 from immutables import Map as ImmutableDict
 
+import shellous
 from shellous.redirect import STDIN_TYPES, STDOUT_APPEND_TYPES, STDOUT_TYPES, Redirect
-from shellous.runner import Runner, run_cmd
+from shellous.runner import Runner
 from shellous.util import coerce_env
 
 # Sentinel used in "mergable" keyword arguments to indicate that a value
@@ -33,7 +36,7 @@ Unset = Union[_T, _UnsetEnum]
 class Options:  # pylint: disable=too-many-instance-attributes
     "Concrete class for per-command options."
 
-    context: "Context" = field(compare=False, repr=False)
+    context: "CmdContext" = field(compare=False, repr=False)
     "Root context object."
 
     env: Optional[ImmutableDict] = field(default=None, repr=False)
@@ -87,6 +90,21 @@ class Options:  # pylint: disable=too-many-instance-attributes
     alt_name: Optional[str] = None
     "Alternate name for the command to use when logging."
 
+    pass_fds: Iterable[int] = ()
+    "File descriptors to pass to the command."
+
+    pass_fds_close: bool = False
+    "True if pass_fds should be closed after subprocess launch."
+
+    write_mode: bool = False
+    "True if using process substitution in write mode."
+
+    start_new_session: bool = False
+    "True if child process should start a new session with setsid()."
+
+    pty: bool = False
+    "True if child process should be controlled using a pseudo-terminal (pty)."
+
     def merge_env(self):
         "Return our `env` merged with the global environment."
         if self.inherit_env:
@@ -151,35 +169,36 @@ class Options:  # pylint: disable=too-many-instance-attributes
 
 
 @dataclass(frozen=True)
-class Context:
+class CmdContext:
     """Concrete class for an immutable execution context."""
 
     options: Options = None  # type: ignore
+    "Default command options."
 
     def __post_init__(self):
         if self.options is None:
             # Initialize `context` in Options to `self`.
             object.__setattr__(self, "options", Options(self))
 
-    def stdin(self, input_, *, close=False):
+    def stdin(self, input_, *, close=False) -> "CmdContext":
         "Return new context with updated `input` settings."
         new_options = self.options.set_stdin(input_, close)
-        return Context(new_options)
+        return CmdContext(new_options)
 
-    def stdout(self, output, *, append=False, close=False):
+    def stdout(self, output, *, append=False, close=False) -> "CmdContext":
         "Return new context with updated `output` settings."
         new_options = self.options.set_stdout(output, append, close)
-        return Context(new_options)
+        return CmdContext(new_options)
 
-    def stderr(self, error, *, append=False, close=False):
+    def stderr(self, error, *, append=False, close=False) -> "CmdContext":
         "Return new context with updated `error` settings."
         new_options = self.options.set_stderr(error, append, close)
-        return Context(new_options)
+        return CmdContext(new_options)
 
-    def env(self, **kwds):
+    def env(self, **kwds) -> "CmdContext":
         """Return new context with augmented environment."""
         new_options = self.options.set_env(kwds)
-        return Context(new_options)
+        return CmdContext(new_options)
 
     def set(  # pylint: disable=unused-argument
         self,
@@ -192,11 +211,16 @@ class Context:
         cancel_timeout=_UNSET,
         cancel_signal=_UNSET,
         alt_name=_UNSET,
-    ):
+        pass_fds=_UNSET,
+        pass_fds_closed=_UNSET,
+        write_mode=_UNSET,
+        start_new_session=_UNSET,
+        pty=_UNSET,
+    ) -> "CmdContext":
         "Return new context with custom options set."
         kwargs = locals()
         del kwargs["self"]
-        return Context(self.options.set(kwargs))
+        return CmdContext(self.options.set(kwargs))
 
     def __call__(self, *args):
         "Construct a new command."
@@ -227,11 +251,13 @@ class Context:
         """
         result = []
         for arg in args:
-            if isinstance(arg, (str, bytes, bytearray, os.PathLike)):
+            if isinstance(
+                arg, (str, bytes, bytearray, os.PathLike, Command, shellous.Pipeline)
+            ):
                 result.append(arg)
             elif isinstance(arg, (list, tuple)):
                 result.extend(self._coerce(arg))
-            elif isinstance(arg, (Command, dict, set)):
+            elif isinstance(arg, (dict, set)):
                 raise NotImplementedError("syntax is reserved")
             elif arg is Ellipsis:
                 raise NotImplementedError("syntax is reserved")
@@ -242,26 +268,13 @@ class Context:
         return tuple(result)
 
 
-def context() -> Context:
-    "Construct a new execution context."
-    return Context()
-
-
-def pipeline(*commands):
-    "Construct a new Pipeline object."
-    # pylint: disable=import-outside-toplevel,cyclic-import
-    from shellous.pipeline import Pipeline
-
-    return Pipeline(commands)
-
-
 @dataclass(frozen=True)
 class Command:
     """A Command instance is lightweight and immutable object that specifies the
     arguments and options used to run a program. Commands do not do anything
     until they are awaited.
 
-    Commands are always created by a Context.
+    Commands are always created by a CmdContext.
 
     ```
     # Create a context.
@@ -275,8 +288,11 @@ class Command:
     ```
     """
 
-    args: Any
+    args: tuple[Union[str, bytes, os.PathLike], ...]
+    "Command arguments including the program name as first argument."
+
     options: Options
+    "Command options."
 
     def __post_init__(self):
         "Validate the command."
@@ -297,34 +313,24 @@ class Command:
             return f"...{name[-31:]}"
         return name
 
-    @property
-    def multiple_capture(self) -> bool:
-        """Return true if the stdin is set to CAPTURE or more than one of
-        stdout, stderr is set to CAPTURE.
-        """
-        return self.options.input == Redirect.CAPTURE or (
-            self.options.output == Redirect.CAPTURE
-            and self.options.error == Redirect.CAPTURE
-        )
-
-    def stdin(self, input_, *, close=False):
+    def stdin(self, input_, *, close=False) -> "Command":
         "Pass `input` to command's standard input."
         new_options = self.options.set_stdin(input_, close)
         return Command(self.args, new_options)
 
-    def stdout(self, output, *, append=False, close=False):
+    def stdout(self, output, *, append=False, close=False) -> "Command":
         "Redirect standard output to `output`."
         _check_args(output, append)
         new_options = self.options.set_stdout(output, append, close)
         return Command(self.args, new_options)
 
-    def stderr(self, error, *, append=False, close=False):
+    def stderr(self, error, *, append=False, close=False) -> "Command":
         "Redirect standard error to `error`."
         _check_args(error, append)
         new_options = self.options.set_stderr(error, append, close)
         return Command(self.args, new_options)
 
-    def env(self, **kwds):
+    def env(self, **kwds) -> "Command":
         """Return new command with augmented environment."""
         new_options = self.options.set_env(kwds)
         return Command(self.args, new_options)
@@ -340,7 +346,12 @@ class Command:
         cancel_timeout: Unset[float] = _UNSET,
         cancel_signal: Unset[Any] = _UNSET,
         alt_name: Unset[Optional[str]] = _UNSET,
-    ):
+        pass_fds: Unset[Iterable[int]] = _UNSET,
+        pass_fds_close: Unset[bool] = _UNSET,
+        write_mode: Unset[bool] = _UNSET,
+        start_new_session: Unset[bool] = _UNSET,
+        pty: Unset[bool] = _UNSET,
+    ) -> "Command":
         """Return new command with custom options set.
 
         - Set `inherit_env` to False to prevent the command from inheriting
@@ -360,19 +371,33 @@ class Command:
         - Set `alt_name` to an alternative name of the command to be displayed
         in logs. Used to resolve ambiguity when the actual command name is a
         scripting language.
+        - Set `pass_fds` to pass open file descriptors to the command.
+        - Set `pass_fds_close` to True to auto-close the `pass_fds`.
+        - Set `write_mode` to True when using process substitution for writing.
+        - Set `start_new_session` to True to start a new session.
+        - Set `pty` to True to use a pseudo-terminal (pty) to control the child
+        process. This setting forces `start_new_session` to
+        also be True.
         """
         kwargs = locals()
         del kwargs["self"]
         return Command(self.args, self.options.set(kwargs))
 
-    def task(self, *, _run_future=None):
-        "Wrap the command in a new asyncio task."
-        return asyncio.create_task(
-            run_cmd(self, _run_future=_run_future),
-            name=f"{self.name}-{id(self)}",
-        )
+    def _replace_args(self, new_args):
+        """Return new command with arguments replaced by `new_args`.
 
-    def run(self):
+        Arguments are NOT type-checked by the context. Program name must be the
+        exact same object.
+        """
+        assert new_args
+        assert new_args[0] is self.args[0]
+        return Command(tuple(new_args), self.options)
+
+    def coro(self, *, _run_future=None):
+        "Return coroutine object to run awaitable."
+        return Runner.run_command(self, _run_future=_run_future)
+
+    def run(self) -> Runner:
         """Return a `Runner` to run the process incrementally.
 
         ```
@@ -386,7 +411,7 @@ class Command:
 
     def __await__(self):
         "Run process and return the standard output."
-        return run_cmd(self).__await__()
+        return self.coro().__await__()
 
     def __call__(self, *args):
         "Apply more arguments to the end of the command."
@@ -406,7 +431,7 @@ class Command:
         "Bitwise or operator is used to build pipelines."
         if isinstance(rhs, STDOUT_TYPES):
             return self.stdout(rhs)
-        return pipeline(self) | rhs
+        return shellous.Pipeline.create(self) | rhs
 
     def __ror__(self, lhs):
         "Bitwise or operator is used to build pipelines."
@@ -420,6 +445,10 @@ class Command:
             return self.stdout(rhs, append=True)
         return NotImplemented
 
+    def __invert__(self):
+        "Unary ~ operator sets write_mode to True."
+        return self.set(write_mode=True)
+
 
 def _check_args(out, append):
     if append and not isinstance(out, STDOUT_APPEND_TYPES):
