@@ -1,21 +1,23 @@
 "Implements support for Pipelines."
 
-import asyncio
 import dataclasses
-import os
 from dataclasses import dataclass
-from typing import Any
 
-from shellous.command import Command
+import shellous
 from shellous.redirect import STDIN_TYPES, STDOUT_APPEND_TYPES, STDOUT_TYPES
-from shellous.runner import PipeRunner, run_pipe
+from shellous.runner import PipeRunner
 
 
 @dataclass(frozen=True)
 class Pipeline:
     "A Pipeline is a sequence of commands."
 
-    commands: Any = ()
+    commands: tuple[shellous.Command, ...] = ()
+
+    @staticmethod
+    def create(*commands) -> "Pipeline":
+        "Create a new Pipeline."
+        return Pipeline(commands)
 
     def __post_init__(self):
         "Validate the pipeline."
@@ -28,39 +30,39 @@ class Pipeline:
         return "|".join(cmd.name for cmd in self.commands)
 
     @property
-    def options(self):
+    def options(self) -> shellous.Options:
         "Return last command's options."
         return self.commands[-1].options
 
-    @property
-    def multiple_capture(self):
-        "Return true if pipe uses multiple capture."
-        return any(cmd.multiple_capture for cmd in self.commands)
-
-    def stdin(self, input_):
+    def stdin(self, input_, *, close=False) -> "Pipeline":
         "Set stdin on the first command of the pipeline."
-        if not self.commands:
-            raise ValueError("invalid pipeline")
-        new_first = self.commands[0].stdin(input_)
+        new_first = self.commands[0].stdin(input_, close=close)
         new_commands = (new_first,) + self.commands[1:]
         return dataclasses.replace(self, commands=new_commands)
 
-    def stdout(self, output, *, append=False):
+    def stdout(self, output, *, append=False, close=False) -> "Pipeline":
         "Set stdout on the last command of the pipeline."
-        if not self.commands:
-            raise ValueError("invalid pipeline")
-        new_last = self.commands[-1].stdout(output, append=append)
+        new_last = self.commands[-1].stdout(output, append=append, close=close)
         new_commands = self.commands[0:-1] + (new_last,)
         return dataclasses.replace(self, commands=new_commands)
 
-    def task(self):
-        "Wrap the command in a new asyncio task."
-        return asyncio.create_task(
-            run_pipe(self),
-            name=f"{self.name}-{id(self)}",
-        )
+    def stderr(self, error, *, append=False, close=False) -> "Pipeline":
+        "Set stderr on the last command of the pipeline."
+        new_last = self.commands[-1].stderr(error, append=append, close=close)
+        new_commands = self.commands[0:-1] + (new_last,)
+        return dataclasses.replace(self, commands=new_commands)
 
-    def run(self):
+    def _set_write_mode(self):
+        "Set write_mode=True on last command of the pipeline."
+        new_last = self.commands[-1].set(write_mode=True)
+        new_commands = self.commands[0:-1] + (new_last,)
+        return dataclasses.replace(self, commands=new_commands)
+
+    def coro(self):
+        "Return coroutine object for pipeline."
+        return PipeRunner.run_pipeline(self)
+
+    def run(self) -> "PipeRunner":
         """Return a `Runner` to help run the pipeline incrementally.
 
         ```
@@ -73,7 +75,7 @@ class Pipeline:
         return PipeRunner(self, capturing=True)
 
     def _add(self, item):
-        if isinstance(item, Command):
+        if isinstance(item, shellous.Command):
             return dataclasses.replace(self, commands=self.commands + (item,))
         if isinstance(item, Pipeline):
             return dataclasses.replace(
@@ -98,7 +100,7 @@ class Pipeline:
         return context._pipe_apply(self, args)
 
     def __or__(self, rhs):
-        if isinstance(rhs, (Command, Pipeline)):
+        if isinstance(rhs, (shellous.Command, Pipeline)):
             return self._add(rhs)
         if isinstance(rhs, STDOUT_TYPES):
             return self.stdout(rhs)
@@ -114,5 +116,9 @@ class Pipeline:
             return self.stdout(rhs, append=True)
         return NotImplemented
 
+    def __invert__(self):
+        "Set write_mode=True option on last command of pipeline."
+        return self._set_write_mode()
+
     def __await__(self):
-        return run_pipe(self).__await__()
+        return self.coro().__await__()
