@@ -5,12 +5,14 @@ import io
 import os
 import sys
 
+import shellous
 import shellous.redirect as redir
+from shellous import pty_util
 from shellous.harvest import harvest, harvest_results
-from shellous.log import LOGGER, log_method
+from shellous.log import LOGGER, log_method, log_timer
 from shellous.redirect import Redirect
 from shellous.result import Result, make_result
-from shellous.util import close_fds, decode
+from shellous.util import close_fds
 
 _NORMAL_LOGGING = True
 _DETAILED_LOGGING = True
@@ -18,11 +20,25 @@ _DETAILED_LOGGING = True
 _KILL_TIMEOUT = 3.0
 _CLOSE_TIMEOUT = 0.25
 
+_KILL_EXIT_CODE = -9 if sys.platform != "win32" else 1
+
 
 def _is_cancelled(ex):
     return isinstance(ex, asyncio.CancelledError)
 
 
+def _is_cmd(cmd):
+    return isinstance(cmd, (shellous.Command, shellous.Pipeline))
+
+
+def _is_write_mode(cmd):
+    "Return true if command/pipeline has write_mode set."
+    if isinstance(cmd, shellous.Pipeline):
+        # Pipelines need to check both the last/first commands.
+        return cmd.options.write_mode or cmd[0].options.write_mode
+    return cmd.options.write_mode
+
+
 class _RunOptions:
     """_RunOptions is context manager to assist in running a command.
 
@@ -43,6 +59,8 @@ class _RunOptions:
         self.input_bytes = None
         self.args = None
         self.kwd_args = None
+        self.subcmds = []
+        self.pty_fds = None
 
     def close_fds(self):
         "Close all open file descriptors in `open_fds`."
@@ -51,7 +69,8 @@ class _RunOptions:
     def __enter__(self):
         "Set up I/O redirections."
         try:
-            self._setup()
+            self._setup_proc_sub()
+            self._setup_redirects()
             return self
         except Exception as ex:
             LOGGER.warning("_RunOptions.enter %r ex=%r", self.command.name, ex)
@@ -65,7 +84,41 @@ class _RunOptions:
                 "_RunOptions.exit %r exc_value=%r", self.command.name, exc_value
             )
 
-    def _setup(self):
+    def _setup_proc_sub(self):
+        "Set up process substitution."
+        if not any(_is_cmd(arg) for arg in self.command.args):
+            return
+
+        if sys.platform == "win32":
+            raise RuntimeError("process substitution not supported on Windows")
+
+        new_args = []
+        pass_fds = []
+
+        for arg in self.command.args:
+            if not _is_cmd(arg):
+                new_args.append(arg)
+                continue
+
+            (read_fd, write_fd) = os.pipe()
+            if _is_write_mode(arg):
+                new_args.append(f"/dev/fd/{write_fd}")
+                pass_fds.append(write_fd)
+                subcmd = arg.stdin(read_fd, close=True)
+            else:
+                new_args.append(f"/dev/fd/{read_fd}")
+                pass_fds.append(read_fd)
+                subcmd = arg.stdout(write_fd, close=True)
+
+            self.subcmds.append(subcmd)
+
+        # pylint: disable=protected-access
+        self.command = self.command._replace_args(new_args).set(
+            pass_fds=pass_fds,
+            pass_fds_close=True,
+        )
+
+    def _setup_redirects(self):
         "Set up I/O redirections."
         options = self.command.options
 
@@ -87,6 +140,18 @@ class _RunOptions:
             sys.stderr,
         )
 
+        # Set up PTY here. This is the first half. Second half in `Runner`.
+        start_new_session = options.start_new_session
+        preexec_fn = None
+        if options.pty:
+            stdin, stdout, stderr, preexec_fn = self._setup_pty1(
+                stdin,
+                stdout,
+                stderr,
+                options.pty,
+            )
+            start_new_session = True
+
         self.input_bytes = input_bytes
         self.args = self.command.args
         self.kwd_args = {
@@ -94,8 +159,15 @@ class _RunOptions:
             "stdout": stdout,
             "stderr": stderr,
             "env": options.merge_env(),
+            "start_new_session": start_new_session,
+            "preexec_fn": preexec_fn,
         }
 
+        if options.pass_fds:
+            self.kwd_args["pass_fds"] = options.pass_fds
+            if options.pass_fds_close:
+                self.open_fds.extend(options.pass_fds)
+
     def _setup_input(self, input_, close, encoding):
         "Set up process input."
         assert input_ is not None
@@ -161,6 +233,40 @@ class _RunOptions:
 
         return stdout
 
+    def _setup_pty1(self, stdin, stdout, stderr, pty):
+        """Set up pseudo-terminal and return (stdin, stdout, stderr, preexec_fn).
+
+        Initializes `self.pty_fds`.
+        """
+
+        parent_fd, child_fd = pty_util.open_pty()
+
+        if stdin == asyncio.subprocess.PIPE:
+            stdin = child_fd
+
+        if stdout == asyncio.subprocess.PIPE:
+            stdout = child_fd
+
+        if stderr == asyncio.subprocess.STDOUT:
+            stderr = child_fd
+        elif stderr == asyncio.subprocess.PIPE:
+            raise RuntimeError("pty can't separate stderr from stdout")
+
+        # If pty is a callable, call it here with `child_fd` as argument. This
+        # gives the client an opportunity to configure the tty.
+        if callable(pty):
+            pty(child_fd)
+
+        self.pty_fds = pty_util.PtyFds(
+            parent_fd,
+            child_fd,
+            pty_util.get_eof(child_fd),
+        )
+
+        LOGGER.info("_setup_pty1: %r", self.pty_fds)
+
+        return stdin, stdout, stderr, lambda: pty_util.set_ctty(child_fd)
+
 
 class Runner:
     """Runner is an asynchronous context manager that runs a command.
@@ -191,12 +297,15 @@ class Runner:
         "Return the command being run."
         return self.options.command
 
-    def result(self, output_bytes=None):
+    def result(self, output_bytes=b""):
         "Check process exit code and raise a ResultError if necessary."
         if self.proc:
             code = self.proc.returncode
         else:
-            code = None
+            # The process was started but cancelled immediately; `self.proc`
+            # was never returned.
+            assert self.cancelled
+            code = _KILL_EXIT_CODE
 
         result = Result(
             output_bytes,
@@ -215,6 +324,7 @@ class Runner:
             task_name = self.name
         task = asyncio.create_task(coro, name=task_name)
         self.tasks.append(task)
+        return task
 
     @log_method(_DETAILED_LOGGING)
     async def _wait(self):
@@ -305,18 +415,27 @@ class Runner:
         try:
             # Set up subprocess arguments and launch subprocess.
             with self.options as opts:
-                self.proc = await asyncio.create_subprocess_exec(
-                    *opts.args,
-                    **opts.kwd_args,
-                )
+                # Launch the main subprocess.
+                with log_timer("asyncio.create_subprocess_exec"):
+                    self.proc = await asyncio.create_subprocess_exec(
+                        *opts.args,
+                        **opts.kwd_args,
+                    )
 
-            # Add a task to monitor for when the process finishes.
-            self.add_task(self.proc.wait(), "proc.wait")
+                # Launch the process substitution commands (if any).
+                for cmd in opts.subcmds:
+                    self.add_task(cmd.coro(), "procsub")
 
             stdin = self.proc.stdin
             stdout = self.proc.stdout
             stderr = self.proc.stderr
 
+            # Second half of pty setup.
+            if opts.pty_fds:
+                assert (stdin, stdout, stderr) == (None, None, None)
+                opts.pty_fds = await opts.pty_fds.open_streams()
+                stdin, stdout = opts.pty_fds.writer, opts.pty_fds.reader
+
             if stderr is not None:
                 error = opts.command.options.error
                 stderr = self._setup_output_sink(stderr, error, opts.encoding, "stderr")
@@ -329,7 +448,11 @@ class Runner:
 
             if stdin is not None:
                 if opts.input_bytes is not None:
-                    self.add_task(redir.write_stream(opts.input_bytes, stdin), "stdin")
+                    eof = opts.pty_fds.eof if opts.pty_fds else None
+                    self.add_task(
+                        redir.write_stream(opts.input_bytes, stdin, eof),
+                        "stdin",
+                    )
                     stdin = None
 
         except (Exception, asyncio.CancelledError) as ex:
@@ -346,11 +469,23 @@ class Runner:
         self.stdout = stdout
         self.stderr = stderr
 
+        # Add a task to monitor for when the process finishes.
+        self.add_task(self._waiter(), "waiter")
+
         return self
 
+    @log_method(_DETAILED_LOGGING)
+    async def _waiter(self):
+        "Run task that waits for process to exit."
+        await self.proc.wait()
+        if self.options.pty_fds:
+            self.stdout._transport.close()  # pylint: disable=protected-access
+
     def _setup_output_sink(self, stream, sink, encoding, tag):
         "Set up a task to write to custom output sink."
         if isinstance(sink, io.StringIO):
+            if encoding is None:
+                raise TypeError("StringIO not supported when encoding=None")
             self.add_task(redir.copy_stringio(stream, sink, encoding), tag)
             stream = None
         elif isinstance(sink, io.BytesIO):
@@ -395,17 +530,20 @@ class Runner:
         "Make sure that our resources are properly closed."
         assert self.proc
 
+        if self.options.pty_fds:
+            self.options.pty_fds.close()
+
         # _close can be called when unwinding exceptions. We need to handle
         # the case that the process has not exited yet. Remember to close the
         # transport.
         if self.proc.returncode is None:
             LOGGER.critical("Runner._close process still running %r", self.proc)
-            self.proc._transport.close()
+            self.proc._transport.close()  # pylint: disable=protected-access
             return
 
         try:
             # Make sure the transport is closed (for asyncio and uvloop).
-            self.proc._transport.close()
+            self.proc._transport.close()  # pylint: disable=protected-access
 
             # Make sure that original stdin is properly closed. `wait_closed`
             # will raise a BrokenPipeError if not all input was properly written.
@@ -414,6 +552,7 @@ class Runner:
                 await harvest(
                     self.proc.stdin.wait_closed(),
                     timeout=_CLOSE_TIMEOUT,
+                    cancel_finish=True,  # finish `wait_closed` if cancelled
                     trustee=self,
                 )
 
@@ -434,16 +573,38 @@ class Runner:
         if self.stdin or (self.stdout and self.stderr):
             raise RuntimeError("multiple capture not supported in iterator")
 
-        encoding = self.options.encoding
         stream = self.stdout or self.stderr
         if stream:
-            async for line in stream:
-                yield decode(line, encoding)
+            async for line in redir.read_lines(stream, self.options.encoding):
+                yield line
 
     def __aiter__(self):
         "Return asynchronous iterator over stdout/stderr."
         return self._readlines()
 
+    @staticmethod
+    async def run_command(command, *, _run_future=None):
+        "Run a command. This is the main entry point for Runner."
+        if not _run_future and _is_multiple_capture(command):
+            LOGGER.warning("run_command: multiple capture requires 'async with'")
+            _cleanup(command)
+            raise ValueError("multiple capture requires 'async with'")
+
+        output_bytes = bytearray()
+
+        async with command.run() as run:
+            if _run_future is not None:
+                # Return streams to caller in another task.
+                _run_future.set_result(run)
+
+            else:
+                # Read the output here and return it.
+                stream = run.stdout or run.stderr
+                if stream:
+                    await redir.copy_bytearray(stream, output_bytes)
+
+        return run.result(bytes(output_bytes))
+
 
 class PipeRunner:  # pylint: disable=too-many-instance-attributes
     """PipeRunner is an asynchronous context manager that runs a pipeline.
@@ -463,10 +624,10 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
 
         self.pipe = pipe
         self.cancelled = False
-        self.tasks = None
+        self.tasks = []
         self.results = None
         self.capturing = capturing
-        self.encoding = pipe.commands[-1].options.encoding
+        self.encoding = pipe.options.encoding
         self.stdin = None
         self.stdout = None
         self.stderr = None
@@ -480,6 +641,16 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
         "Return `Result` object for PipeRunner."
         return make_result(self.pipe, self.results, self.cancelled)
 
+    def add_task(self, coro, tag=None):
+        "Add a background task."
+        if tag:
+            task_name = f"{self.name}#{tag}"
+        else:
+            task_name = self.name
+        task = asyncio.create_task(coro, name=task_name)
+        self.tasks.append(task)
+        return task
+
     @log_method(_DETAILED_LOGGING)
     async def _wait(self, *, kill=False):
         "Wait for pipeline to finish."
@@ -544,7 +715,8 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
             if self.capturing:
                 stdin, stdout, stderr = await self._setup_capturing(cmds)
             else:
-                self.tasks = [cmd.task() for cmd in cmds]
+                for cmd in cmds:
+                    self.add_task(cmd.coro())
 
             self.stdin = stdin
             self.stdout = stdout
@@ -555,6 +727,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
         except BaseException:  # pylint: disable=broad-except
             # Clean up after any exception *including* CancelledError.
             close_fds(open_fds)
+            raise
 
     def _setup_pipeline(self, open_fds):
         """Return the pipeline stitched together with pipe fd's.
@@ -584,11 +757,15 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
         loop = asyncio.get_event_loop()
         first_fut = loop.create_future()
         last_fut = loop.create_future()
-        first_task = cmds[0].task(_run_future=first_fut)
-        last_task = cmds[-1].task(_run_future=last_fut)
 
-        middle_tasks = [cmd.task() for cmd in cmds[1:-1]]
-        self.tasks = [first_task] + middle_tasks + [last_task]
+        first_coro = cmds[0].coro(_run_future=first_fut)
+        last_coro = cmds[-1].coro(_run_future=last_fut)
+        middle_coros = [cmd.coro() for cmd in cmds[1:-1]]
+
+        self.add_task(first_coro)
+        for coro in middle_coros:
+            self.add_task(coro)
+        self.add_task(last_coro)
 
         # When capturing, we need the first and last commands in the
         # pipe to signal when they are ready.
@@ -617,42 +794,47 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
         if self.stdin or (self.stdout and self.stderr):
             raise RuntimeError("multiple capture not supported in iterator")
 
-        encoding = self.encoding
         stream = self.stdout or self.stderr
         if stream:
-            async for line in stream:
-                yield decode(line, encoding)
+            async for line in redir.read_lines(stream, self.encoding):
+                yield line
 
     def __aiter__(self):
         "Return asynchronous iterator over stdout/stderr."
         return self._readlines()
 
+    @staticmethod
+    async def run_pipeline(pipe):
+        "Run a pipeline. This is the main entry point for PipeRunner."
 
-async def run_cmd(command, *, _run_future=None):
-    "Run a command."
-    if not _run_future and command.multiple_capture:
-        raise ValueError("multiple capture requires 'async with'")
-
-    output_bytes = None
-
-    async with command.run() as run:
-        if _run_future is not None:
-            # Return streams to caller in another task.
-            _run_future.set_result(run)
-
-        else:
-            # Read the output here and return it.
-            stream = run.stdout or run.stderr
-            if stream:
-                output_bytes = await stream.read()
-
-    return run.result(output_bytes)
+        run = PipeRunner(pipe, capturing=False)
+        async with run:
+            pass
+        return run.result()
 
 
-async def run_pipe(pipe):
-    "Run a pipeline"
+def _is_multiple_capture(cmd):
+    "Return true if stdin is CAPTURE or both stdout and stderr are CAPTURE."
+    return cmd.options.input == Redirect.CAPTURE or (
+        cmd.options.output == Redirect.CAPTURE and cmd.options.error == Redirect.CAPTURE
+    )
 
-    run = PipeRunner(pipe, capturing=False)
-    async with run:
-        pass
-    return run.result()
+
+def _cleanup(command):
+    "Close remaining file descriptors that need to be closed."
+
+    def _add_close(close, fdesc):
+        if close:
+            if isinstance(fdesc, (int, io.IOBase)):
+                open_fds.append(fdesc)
+
+    open_fds = []
+
+    _add_close(command.options.input_close, command.options.input)
+    _add_close(command.options.output_close, command.options.output)
+    _add_close(command.options.error_close, command.options.error)
+
+    if command.options.pass_fds_close:
+        open_fds.extend(command.options.pass_fds)
+
+    close_fds(open_fds)
