@@ -1,21 +1,20 @@
 "Implements various utility functions."
 
+import io
 import os
-from typing import Optional, Union
+from typing import Any, Optional, Union
+
+from .log import LOGGER
 
 
-def decode(data: Optional[bytes], encoding: Optional[str]) -> Union[str, bytes]:
+def decode(data: Optional[bytes], encoding: str) -> str:
     "Utility function to decode optional byte strings."
-    if encoding is None:
-        if data is None:
-            return b""
-        return data
     if not data:
         return ""
     return data.decode(*encoding.split(maxsplit=1))
 
 
-def coerce_env(env: dict):
+def coerce_env(env: dict[str, Any]) -> dict[str, str]:
     """Utility function to coerce environment variables to string.
 
     If the value of an environment variable is `...`, grab the value from the
@@ -30,7 +29,7 @@ def coerce_env(env: dict):
     return {str(key): _coerce(key, value) for key, value in env.items()}
 
 
-def close_fds(open_fds):
+def close_fds(open_fds: list[Union[io.IOBase, int]]) -> None:
     "Close open file descriptors or file objects."
     try:
         for obj in open_fds:
@@ -38,8 +37,8 @@ def close_fds(open_fds):
                 if obj >= 0:
                     try:
                         os.close(obj)
-                    except OSError:
-                        pass
+                    except OSError as ex:
+                        LOGGER.warning("os.close ex=%r", ex)
             else:
                 obj.close()
     finally:
